/*	$OpenBSD: autoconf.c,v 1.7 2000/03/03 00:54:53 todd Exp $	*/
/*
 * Copyright (c) 1998 Steve Murphree, Jr.
 * Copyright (c) 1996 Nivas Madhur
 * Copyright (c) 1994 Christian E. Hopps
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Christian E. Hopps.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/map.h>
#include <sys/buf.h>
#include <sys/dkstat.h>
#include <sys/dmap.h>
#include <sys/reboot.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/disklabel.h>
#include <machine/vmparam.h>
#include <machine/asm_macro.h>   /* enable/disable interrupts */
#include <machine/cpu.h>
#include <machine/autoconf.h>
#include <machine/disklabel.h>
#include <machine/pte.h>

/*
 * The following several variables are related to
 * the configuration process, and are used in initializing
 * the machine.
 */

struct   device *parsedisk __P((char *, int, int, dev_t *));
void  setroot __P((void));
void  swapconf __P((void));
void  configure __P((void));
char    buginchr __P((void));
int   getsb __P((char *, int));

int cold;   /* 1 if still booting */
#include <sys/kernel.h>

/* XXX must be allocated statically because of early console init */
struct   map extiomap[EIOMAPSIZE/16];
extern   void *extiobase;

/*
 * called at boot time, configure all devices on the system.
 */
void
configure()
{
   bootdv = NULL; /* set by device drivers (if found) */
/*	rminit(extiomap, (long)EIOMAPSIZE, (long)1, "extio", EIOMAPSIZE/16);*/

   if (config_rootfound("mainbus", "mainbus") == 0)
      panic("no mainbus found");

   /*
    * Turn external interrupts on. We have all the drivers in
    * place now!
    */
   enable_interrupt();
   spl0();
   setroot();
   swapconf();
   /*
    * Done with autoconfig!
    */
   cold = 0;
}

/*ARGSUSED*/
int
simple_devprint(auxp, pnp)
void *auxp;
char *pnp;
{
   return (QUIET);
}

int
matchname(fp, sp)
char *fp, *sp;
{
   int len;

   len = strlen(fp);
   if (strlen(sp) != len)
      return (0);
   if (bcmp(fp, sp, len) == 0)
      return (1);
   return (0);
}

/*
 * Configure swap space and related parameters.
 */
void 
swapconf()
{
   register struct swdevt *swp;
   register int nblks;

   for (swp = swdevt; swp->sw_dev != NODEV; swp++)
      if (bdevsw[major(swp->sw_dev)].d_psize) {
         nblks =
         (*bdevsw[major(swp->sw_dev)].d_psize)(swp->sw_dev);
         if (nblks != -1 &&
             (swp->sw_nblks == 0 || swp->sw_nblks > nblks))
            swp->sw_nblks = nblks;
      }
   dumpconf();
}

/*
 * the rest of this file was adapted from Theo de Raadt's code in the 
 * sparc port to nuke the "options GENERIC" stuff.
 */

struct nam2blk {
   char *name;
   int maj;
} nam2blk[] = {
   { "sd",     4},
   { "st",     5},
   { "rd",     7},
};


static int
findblkmajor(dv)
struct device *dv;
{
   char *name = dv->dv_xname;
   register int i;

   for (i = 0; i < sizeof(nam2blk)/sizeof(nam2blk[0]); ++i)
      if (strncmp(name, nam2blk[i].name, strlen(nam2blk[0].name)) == 0)
         return (nam2blk[i].maj);
   return (-1);
}

static struct device *
getdisk(str, len, defpart, devp)
char *str;
int len, defpart;
dev_t *devp;
{
   register struct device *dv;

   if ((dv = parsedisk(str, len, defpart, devp)) == NULL) {
      printf("use one of:");
      for (dv = alldevs.tqh_first; dv != NULL;
          dv = dv->dv_list.tqe_next) {
         if (dv->dv_class == DV_DISK)
            printf(" %s[a-h]", dv->dv_xname);
#ifdef NFSCLIENT
         if (dv->dv_class == DV_IFNET)
            printf(" %s", dv->dv_xname);
#endif
      }
      printf("\n");
   }
   return (dv);
}

struct device *
parsedisk(str, len, defpart, devp)
char *str;
int len, defpart;
dev_t *devp;
{
   register struct device *dv;
   register char *cp, c;
   int majdev, mindev, part;

   if (len == 0)
      return (NULL);
   cp = str + len - 1;
   c = *cp;
   if (c >= 'a' && c <= 'h') {
      part = c - 'a';
      *cp = '\0';
   } else
      part = defpart;

   for (dv = alldevs.tqh_first; dv != NULL; dv = dv->dv_list.tqe_next) {
      if (dv->dv_class == DV_DISK &&
          strcmp(str, dv->dv_xname) == 0) {
         majdev = findblkmajor(dv);
         if (majdev < 0)
            panic("parsedisk");
         mindev = (dv->dv_unit << PARTITIONSHIFT) + part;
         *devp = makedev(majdev, mindev);
         break;
      }
#ifdef NFSCLIENT
      if (dv->dv_class == DV_IFNET &&
          strcmp(str, dv->dv_xname) == 0) {
         *devp = NODEV;
         break;
      }
#endif
   }

   *cp = c;
   return (dv);
}

/*
 * Attempt to find the device from which we were booted.
 * If we can do so, and not instructed not to do so,
 * change rootdev to correspond to the load device.
 *
 * XXX Actually, swap and root must be on the same type of device,
 * (ie. DV_DISK or DV_IFNET) because of how (*mountroot) is written.
 * That should be fixed.
 */
void
setroot()
{
   register struct swdevt *swp;
   register struct device *dv;
   register int len, majdev, mindev;
   dev_t nrootdev, nswapdev = NODEV;
   char buf[128];
   dev_t temp;
#if defined(NFSCLIENT)
   extern char *nfsbootdevname;
#endif

   printf("boot device: %s\n",
          (bootdv) ? bootdv->dv_xname : "<unknown>");

   if (boothowto & RB_ASKNAME) {
      for (;;) {
         printf("root device");
         if (bootdv != NULL)
            printf("(default %s%c)",
                   bootdv->dv_xname,
                   bootdv->dv_class == DV_DISK
                   ? 'a' : ' ');
         printf(": ");
         len = getsb(buf, sizeof(buf));
         if (len == 0 && bootdv != NULL) {
            strcpy(buf, bootdv->dv_xname);
            len = strlen(buf);
         }
         if (len > 0 && buf[len - 1] == '*') {
            buf[--len] = '\0';
            dv = getdisk(buf, len, 1, &nrootdev);
            if (dv != NULL) {
               bootdv = dv;
               nswapdev = nrootdev;
               goto gotswap;
            }
         }
         dv = getdisk(buf, len, 0, &nrootdev);
         if (dv != NULL) {
            bootdv = dv;
            break;
         }
      }

      /*
       * because swap must be on same device as root, for
       * network devices this is easy.
       */
      if (bootdv->dv_class == DV_IFNET) {
         goto gotswap;
      }
      for (;;) {
         printf("swap device ");
         if (bootdv != NULL)
            printf("(default %s%c)",
                   bootdv->dv_xname,
                   bootdv->dv_class == DV_DISK?'b':' ');
         printf(": ");
         len = getsb(buf, sizeof(buf));
         if (len == 0 && bootdv != NULL) {
            switch (bootdv->dv_class) {
               case DV_IFNET:
                  nswapdev = NODEV;
                  break;
               case DV_DISK:
                  nswapdev = makedev(major(nrootdev),
                                     (minor(nrootdev) & ~ PARTITIONMASK) | 1);
                  break;
               case DV_TAPE:
               case DV_TTY:
               case DV_DULL:
               case DV_CPU:
                  break;
            }
            break;
         }
         dv = getdisk(buf, len, 1, &nswapdev);
         if (dv) {
            if (dv->dv_class == DV_IFNET)
               nswapdev = NODEV;
            break;
         }
      }
      gotswap:
      rootdev = nrootdev;
      dumpdev = nswapdev;
      swdevt[0].sw_dev = nswapdev;
      swdevt[1].sw_dev = NODEV;

   } else if (mountroot == NULL) {

      /*
       * `swap generic': Use the device the ROM told us to use.
       */
      if (bootdv == NULL)
         panic("boot device not known");

      majdev = findblkmajor(bootdv);
      if (majdev >= 0) {
         /*
          * Root and swap are on a disk.
          * val[2] of the boot device is the partition number.
          * Assume swap is on partition b.
          */
         int part = bootpart;
         mindev = (bootdv->dv_unit << PARTITIONSHIFT) + part;
         rootdev = makedev(majdev, mindev);
         nswapdev = dumpdev = makedev(major(rootdev),
                                      (minor(rootdev) & ~ PARTITIONMASK) | 1);
      } else {
         /*
          * Root and swap are on a net.
          */
         nswapdev = dumpdev = NODEV;
      }
      swdevt[0].sw_dev = nswapdev;
      swdevt[1].sw_dev = NODEV;

   } else {

      /*
       * `root DEV swap DEV': honour rootdev/swdevt.
       * rootdev/swdevt/mountroot already properly set.
       */
      return;
   }

   switch (bootdv->dv_class) {
#if defined(NFSCLIENT)
      case DV_IFNET:
         mountroot = nfs_mountroot;
         nfsbootdevname = bootdv->dv_xname;
         return;
#endif
#if defined(FFS)
      case DV_DISK:
         mountroot = dk_mountroot;
         majdev = major(rootdev);
         mindev = minor(rootdev);
         printf("root on %s%c\n", bootdv->dv_xname,
                (mindev & PARTITIONMASK) + 'a');
         break;
#endif
      default:
         printf("can't figure root, hope your kernel is right\n");
         return;
   }

   /*
    * XXX: What is this doing?
    */
   mindev &= ~PARTITIONMASK;
   temp = NODEV;
   for (swp = swdevt; swp->sw_dev != NODEV; swp++) {
      if (majdev == major(swp->sw_dev) &&
          mindev == (minor(swp->sw_dev) & ~PARTITIONMASK)) {
         temp = swdevt[0].sw_dev;
         swdevt[0].sw_dev = swp->sw_dev;
         swp->sw_dev = temp;
         break;
      }
   }
   if (swp->sw_dev == NODEV)
      return;

   /*
    * If dumpdev was the same as the old primary swap device, move
    * it to the new primary swap device.
    */
   if (temp == dumpdev)
      dumpdev = swdevt[0].sw_dev;
}

/*
 * find a device matching "name" and unit number
 */
struct device *
getdevunit(name, unit)
char *name;
int unit;
{
   struct device *dev = alldevs.tqh_first;
   char num[10], fullname[16];
   int lunit;

   /* compute length of name and decimal expansion of unit number */
   sprintf(num, "%d", unit);
   lunit = strlen(num);
   if (strlen(name) + lunit >= sizeof(fullname) - 1)
      panic("config_attach: device name too long");

   strcpy(fullname, name);
   strcat(fullname, num);

   while (strcmp(dev->dv_xname, fullname) != 0) {
      if ((dev = dev->dv_list.tqe_next) == NULL)
         return NULL;
   }
   return dev;
}

int
getsb(cp, size)
char *cp;
int size;
{
   register char *lp;
   register int len;
   register int c;

   lp = cp;
   len = 0;
   for (;;) {
      c = buginchr();

      switch (c) {
         case '\n':
         case '\r':
            printf("\n");
            *lp++ = '\0';
            return (len);
         case '\b':
         case '\177':
            if (len) {
               printf("\b \b");
               --lp;
               --len;
            }
            break;
         case 'u' & 037:
            while (len) {
               printf("\b \b");
               --lp;
               --len;
            }
            break;
         case '\t':
            c = ' ';
         default:
            if (len + 1 >= size || c < ' ') {
               printf("\007");
               break;
            }
            printf("%c", c);
            ++len;
            *lp++ = c;
      }
   }
}

/*
 * Slave CPU pre-main routine.
 * Determine CPU number and set it.
 *
 * Running on an interrupt stack here; do nothing fancy.
 *
 * Called from "mvme88k/locore.S"
 */
void slave_pre_main(void)
{
   set_cpu_number(cmmu_cpu_number()); /* Determine cpu number by CMMU */
   splhigh();
   enable_interrupt();
}


